/*
Firmware to detect shadows on a line-CCD from an Epson SX125
(C) 2011 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
Rough description of the protocol to the PC: serial, 115200 baud, n81.
Consists of 16-bit values, high bit is sent first.

One frame:
0xAAFE - start frame
0xAAxx - data for led x follows:
start  - start position of the shadow cast by led X
length - length of the shadow cast by led X
0xAAFF - end frame


There can be any amount (including 0) of the start/lenghth pairs; the end of these is
indicated by receiving a 0xAA??-value. There are multiple led datapackets per frame.
*/


#define F_CPU 20000000
#include <util/delay.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include <string.h>
#include "stdout.h"

#define PIXELS 2500
#define SKIPFIRST 88 //pixels to skip at start of measurement
#define MINSIZE 50 //about half a finger

#define CCDCLK 		PD5
#define CCDCLKIN	PB5
#define CCDDATA		PB4
#define CCDLATCH	PB1


#define LED1	PD2
#define LED2	PD3
#define LED3	PD4
#define LED4	PD7
#define LED5	PD6

char dataIn[PIXELS/8];

//Turn on a certain IR LED
void selLed(char led) {
	if (led==1) PORTD|=(1<<LED1); else PORTD&=~(1<<LED1);
	if (led==2) PORTD|=(1<<LED2); else PORTD&=~(1<<LED2);
	if (led==3) PORTD|=(1<<LED3); else PORTD&=~(1<<LED3);
	if (led==4) PORTD|=(1<<LED4); else PORTD&=~(1<<LED4);
	if (led==5) PORTD|=(1<<LED5); else PORTD&=~(1<<LED5);
}

//Receive a line of data from the CCD
void recvLine(char* buff) {
	int x;
	char d;

	while(TCNT0>1) ;
	PORTB|=(1<<CCDLATCH);
	while(TCNT0>1) ;
	PORTB&=~(1<<CCDLATCH);

	for (x=0; x<(SKIPFIRST/8); x++) {
		while((SPSR&0x80)==0);
		d=SPDR;
	}


	for (x=0; x<(PIXELS/8); x++) {
		while((SPSR&0x80)==0);
		buff[x]=SPDR;
	}
}

//Send a 16-bit value over the serial line.
void serSendInt(int val) {
    while(!(UCSR0A&0x20)) ;
    UDR0=val>>8;
    while(!(UCSR0A&0x20)) ;
    UDR0=val&0xff;
}

//Convert the bit array in data to (start, length)-pairs and send them over the serial line.
void parseGaps(char *data) {
	int y;
	int pixel=0, len=0;
	char x;
	int start=-1;
	for (y=0; y<PIXELS/8; y++) {
		for (x=0x80; x!=0; x>>=1) {
			if (!(data[y]&x)) {
				if (start==-1) start=pixel;
			} else {
				if (start!=-1) {
					len=pixel-start;
					if (len>MINSIZE) {
						serSendInt(start);
						serSendInt(len);
					}
					start=-1;
				}
			}
			pixel++;
		}
	}
	//If there's something to send left, send it.
	if (start!=-1) {
		len=pixel-start;
		if (len>MINSIZE) {
			serSendInt(start);
			serSendInt(len);
		}
	}
}

//Main routine
int main(void) {
	char l=1;
	stdoutInit(10); //20MHz -> 115200 baud

	DDRB=(1<<CCDLATCH);
	DDRD=(1<<CCDCLK)|(1<<LED1)|(1<<LED2)|(1<<LED3)|(1<<LED4)|(1<<LED5);

	TCCR0A=0x23;
	TCCR0B=0x09;
	OCR0A=14;
	OCR0B=14/2;

	SPCR=0x45;
	SPSR=0;
	
	while(1) {
		selLed(l);
		recvLine(dataIn); //first one just flushes the pipeline :/
		recvLine(dataIn);
		selLed(0);
		serSendInt(0xAA00+(int)l);
		parseGaps(dataIn);
		l++;
		if (l==6) {
			serSendInt(0xAAFF);
			serSendInt(0xAAFE);
			l=1;
		}
	}
}
