/*
Software to convert a line-CCD plus a bunch of LEDs to a touchscreen.
(C) 2011 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
	    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
			    
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "calibrate.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "vector.h"
#include "misc.h"

#define CCDMAX (2500-88)


//Calibration data
static VectorStruct ledPosMed[MAXLED];
static int ccdXStart, ccdXEnd;
static int ccdHeightAvg;


//Draws a point, wait till user touched it and returns the locations touched per led, or -1 if none detected
static void getMeasForCalPoint(int fd, cairo_t *cr, int x, int y, int* measret) {
	int n, no;
	BlackMark meas[100], rmeas[100];
	for (n=0; n<MAXLED; n++) measret[n]=-1;

	cairo_set_source_rgb(cr, 255,255,255);
	cairo_paint(cr);
	cairo_set_source_rgb(cr, 0,0,0);
	cairo_move_to(cr, x-10, y);
	cairo_line_to(cr, x+10, y);
	cairo_move_to(cr, x, y-10);
	cairo_line_to(cr, x, y+10);
	cairo_stroke(cr);
	cairoSdlBlit(cr);

	while(serialRead(fd, meas, 100)==0) ;
	//Delay a while.
	for (n=0; n<15; n++) serialRead(fd, meas, 100);

	//Read and display measurements
	no=serialRead(fd, rmeas, 100);
	drawMeasLines(cr, meas, no);
	cairoSdlBlit(cr);

	//Wait till no finger is detected.
	while(serialRead(fd, meas, 100)!=0) ;

	for (n=0; n<no; n++) {
		measret[rmeas[n].led]=rmeas[n].start+rmeas[n].len/2;
	}
}

inline void waitTouch(int fd) {
	BlackMark meas[100];
	while(serialRead(fd, meas, 100)==0) ;
}

#define NOPOINTS 9
//Calibrate the display by asking the user to touch a bunch of points.
void calibrate(int fd, cairo_t *cr) {

	typedef struct {
		VectorStruct pos;
		int meas[MAXLED];
	} PointsStruct;
	PointsStruct points[NOPOINTS];

	typedef struct {
		int led;
		VectorStruct base;
		VectorStruct dir;
		int sensorpos;
	} LineStruct;
	LineStruct lines[10240];

	typedef struct {
		int led;
		VectorStruct pos;
	} LedPosStruct;
	LedPosStruct ledPos[10240];
	int ledPosCnt;
	int ccdHeight[10240];
	int ccdHeightCnt;

	typedef struct {
		int start;
		int end;
	} SensorPosStruct;
	SensorPosStruct sensorPos[10240];
	int sensorPosCnt;

	int min;

	VectorStruct v;

	int x,y,n,ret;
	int p1, p2, p3;
	int lineNo;
	int before, after;

	FILE *f;

	//Generate points to be touched.
	n=0;
	for (x=0; x<3; x++) {
		for (y=0; y<3; y++) {
			points[n].pos.x=x*200+300;
			points[n].pos.y=y*180+400;
			n++;
		}
	}

	f=fopen("calibration", "r");
	if (f==NULL) {
		//Get measurements for points.
		for (n=0; n<NOPOINTS; n++) {
			getMeasForCalPoint(fd, cr, points[n].pos.x, points[n].pos.y, points[n].meas);
			printf("Got: ");
			for (x=0; x<8; x++) printf("%i ", points[n].meas[x]);
			printf("\n");
		}
		f=fopen("calibration", "w");
		for (x=0; x<NOPOINTS; x++) fwrite(points[x].meas, sizeof(int), MAXLED, f);
		fclose(f);
	} else {
		for (x=0; x<NOPOINTS; x++) fread(points[x].meas, sizeof(int), MAXLED, f);
		fclose(f);
	}


	//We have points. See if we can derive the positions of the LEDs and line-sensor from that...
	//First, test all triplets of 3 calibrated points against eachother.
	//If they're different, available, 2 are in-line on the x-axis and 2 on the y-axis then
	//we can use them.
	lineNo=0;
	for (n=0; n<MAXLED; n++) {
		for (p1=0; p1<NOPOINTS; p1++) {
			for (p2=0; p2<NOPOINTS; p2++) {
				for (p3=0; p3<NOPOINTS; p3++) {
					if (p1!=p2 && p2!=p3 &&
					 points[p1].meas[n]!=-1 &&
					 points[p2].meas[n]!=-1 &&
					 points[p3].meas[n]!=-1 &&
					 points[p1].pos.x!=points[p2].pos.x && points[p2].pos.x==points[p3].pos.x &&
					 points[p1].pos.y==points[p2].pos.y && points[p2].pos.y!=points[p3].pos.y) {
//						printf("Constructing line: led=%i p1=%i, p2=%i, p3=%i\n",n, p1, p2, p3);
						//We now have something like:
						//
						//             p3
						//             |
						//     p1------p2
						//
						//with p1-p2 being perfectly horizontal and p2-p3 being perfectly vertical.
						//From this and the measurements, we can calculate the direction from P3 to the illuminated LED.
						lines[lineNo].led=n;
						lines[lineNo].sensorpos=points[p3].meas[n];
						lines[lineNo].base.x=points[p3].pos.x;
						lines[lineNo].base.y=points[p3].pos.y;
						lines[lineNo].dir.y=(points[p2].pos.y-points[p3].pos.y)*(points[p2].meas[n]-points[p1].meas[n]);
						lines[lineNo].dir.x=(points[p2].pos.x-points[p1].pos.x)*(points[p3].meas[n]-points[p2].meas[n]);
						if (lines[lineNo].dir.y>0) { //make sure vector points up
							lines[lineNo].dir.y=-lines[lineNo].dir.y;
							lines[lineNo].dir.x=-lines[lineNo].dir.x;
						}
						lineNo++;
					}
				}
			}
		}
	}

	printf("Calculated %i lines.\n",lineNo);

	//Calculate where the lines in the direction of the LEDs intersect.
	ledPosCnt=0;
	for (p1=0; p1<lineNo; p1++) {
		for (p2=p1; p2<lineNo; p2++) {
			if (p1!=p2 && lines[p1].led==lines[p2].led) {
				ret=getIntersectPoint(lines[p1].base, lines[p1].dir, lines[p2].base, lines[p2].dir, &v);
	//			printf("Led %i: %i and %i %sintersect at %f,%f.\n", lines[p1].led, p1, p2, ret?"":"don't ", v.x, v.y);
				//Add only if led is above screen and lines aren't parallel.
				if (ret && v.y<0) {
					ledPos[ledPosCnt].led=lines[p1].led;
					ledPos[ledPosCnt].pos.x+=v.x;
					ledPos[ledPosCnt].pos.y+=v.y;
					ledPosCnt++;
				}
			}
		}
	}
	printf("Calculated %i intersections.\n",ledPosCnt);

	//Fetch the median of all led points. Sort by Y only... yes, this isn't the best approach but I
	//can't decide on a method to get the median of a bunch of 2d-points... does it even exist?
	for (n=0; n<MAXLED; n++) {
		ledPosMed[n].x=0;
		ledPosMed[n].y=0;
	}
	for (p1=0; p1<ledPosCnt; p1++) {
		before=0; after=0;
		for (p2=0; p2<ledPosCnt; p2++) {
			if (ledPos[p1].led==ledPos[p2].led && p1!=p2) {
			 	if (ledPos[p1].pos.y<ledPos[p2].pos.y) {
					before++; 
				} else {
					after++;
				}
			}
		}
		if (abs(before-after)<=1) {
			ledPosMed[ledPos[p1].led].x=ledPos[p1].pos.x;
			ledPosMed[ledPos[p1].led].y=ledPos[p1].pos.y;
		}
	}
	//Now, the best guesses at the points where the LEDs reside are in ledPosMed.

	//Now, calculate at what height the line ccd is.
	//For this, we need three lines which go through two points which are on the same y-position.
	//ToDo: This needs to be calculated on the measurements we took, not on the lines we derived from those, because
	//the lines are not as precise and not as numerous as raw measurements plus calculated directions.
	ccdHeightCnt=0;
	for (p1=0; p1<lineNo; p1++) {
		for (p2=0; p2<lineNo; p2++) {
			for (p3=0; p3<lineNo; p3++) {
				if (p1!=p2 && p2!=p3 && p2<p1 &&
				 lines[p1].base.x==lines[p2].base.x && lines[p1].base.x!=lines[p3].base.x &&
				 lines[p1].base.y==lines[p2].base.y && lines[p1].base.y==lines[p3].base.y &&
//				 lines[p1].dir.x<0 && lines[p2].dir.x>0 && lines[p3].dir.x<0 &&
				 lines[p1].sensorpos!=lines[p2].sensorpos &&
				 lines[p1].base.x<lines[p3].base.x) {
					float x, l, c;
					int h;
					l=(lines[p3].base.x-lines[p2].base.x);
//					printf("a+b=%i c+d=%i\n", lines[p1].sensorpos-lines[p2].sensorpos, lines[p3].sensorpos-lines[p1].sensorpos);
					x=((double)lines[p1].sensorpos-lines[p2].sensorpos)/((double)lines[p3].sensorpos-lines[p1].sensorpos);
					c=(lines[p1].dir.x/lines[p1].dir.y)*(x+1)-(lines[p2].dir.x/lines[p2].dir.y)+(lines[p3].dir.x/lines[p3].dir.y);
					h=((l*x)/c);
					if (h>0) {
						ccdHeight[ccdHeightCnt]=h+lines[p1].base.y;
//						printf("%i %i %i y=%f x1=%f x2=%f l=%f xdiv=%f  Ccd ypos = %i\n", p1, p2, p3, lines[p1].base.y, lines[p1].base.x, lines[p3].base.x, l, x, ccdHeight[ccdHeightCnt]);
						ccdHeightCnt++;
					}
				}
			}
		}
	}
	printf("CcdHeight array: %i\n", ccdHeightCnt);

	//Take the median of the calculated heights.
	min=99999;
	ccdHeightAvg=0;
	for (p1=0; p1<ccdHeightCnt; p1++) {
		before=0; after=0;
		for (p2=0; p2<ccdHeightCnt; p2++) {
			if (ccdHeight[p1]<ccdHeight[p2]) before++;
			if (ccdHeight[p1]>ccdHeight[p2]) after++;
		}
		if (abs(before-after)<=min) {
			ccdHeightAvg=ccdHeight[p1];
			min=abs(before-after);
		}
	}

	printf("Ccd is at height %i.\n",ccdHeightAvg);

	//Now we only need to find out the locations of the extremities of the CCD.
	sensorPosCnt=0;
	for (p1=0; p1<lineNo; p1++) {
		for (p2=0; p2<lineNo; p2++) {
			if (lines[p1].base.y==lines[p2].base.y &&
			 lines[p1].base.x!=lines[p2].base.x) {
				VectorStruct ccdline, straightdir;
				VectorStruct rsh1,rsh2;
				double scale;
				int start, end;
				//Get length between the two shadows by looking where the two beams arrive at the ccd height
				straightdir.x=1; straightdir.y=0;
				ccdline.x=0; ccdline.y=ccdHeightAvg;
				getIntersectPoint(ccdline, straightdir, lines[p1].base, lines[p1].dir, &rsh1);
				getIntersectPoint(ccdline, straightdir, lines[p2].base, lines[p2].dir, &rsh2);
				scale=(rsh2.x-rsh1.x)/(lines[p2].sensorpos-lines[p1].sensorpos);
				start=rsh1.x-(lines[p1].sensorpos*scale);
				end=rsh1.x+((CCDMAX-lines[p1].sensorpos)*scale);
				sensorPos[sensorPosCnt].start=start;
				sensorPos[sensorPosCnt].end=end;
				sensorPosCnt++;
//				printf("Sensor: %i-%i\n", start, end);
			}
		}
	}
	printf("Sensorpos array: %i\n", sensorPosCnt);

	//Take the median of the sensor length. Presumes the absolute position at that measurement will be OK too.
	min=99999;
	for (p1=0; p1<sensorPosCnt; p1++) {
		before=0; after=0;
		for (p2=0; p2<sensorPosCnt; p2++) {
			if ((sensorPos[p1].end-sensorPos[p1].start)<(sensorPos[p2].end-sensorPos[p2].start)) before++;
			if ((sensorPos[p1].end-sensorPos[p1].start)>(sensorPos[p2].end-sensorPos[p2].start)) after++;
		}
//		printf("%i: %i\n", p1, before-after);
		if (abs(before-after)<=min) {
			ccdXStart=sensorPos[p1].start;
			ccdXEnd=sensorPos[p1].end;
			min=abs(before-after);
		}
	}
	printf("Sensor: from %i to %i\n", ccdXStart, ccdXEnd);


	//Phew, that was something. Show our work.
	if (cr!=NULL) {
		cairo_set_source_rgb(cr, 255,255,255);
		cairo_paint(cr);
		cairo_set_line_width(cr, .7);
		for (n=0; n<lineNo; n++) {
			resizeVect(&lines[n].dir, 500);
//			printf("Led %i: (%f, %f) dir (%f, %f)\n", lines[n].led, lines[n].base.x, lines[n].base.y, lines[n].dir.x, lines[n].dir.y);
			if (lines[n].led==0) cairo_set_source_rgb(cr, 1,0,0);
			if (lines[n].led==1) cairo_set_source_rgb(cr, 0,1,0);
			if (lines[n].led==2) cairo_set_source_rgb(cr, 0,0,1);
			if (lines[n].led==3) cairo_set_source_rgb(cr, 1,0,1);
			if (lines[n].led==4) cairo_set_source_rgb(cr, 1,1,0);
			if (lines[n].led==5) cairo_set_source_rgb(cr, 0,1,1);
			cairo_move_to(cr, lines[n].base.x, lines[n].base.y);
			cairo_rel_line_to(cr, lines[n].dir.x, lines[n].dir.y);
			cairo_stroke(cr);
		}

		cairo_set_line_width(cr, 2);
		for (n=0; n<MAXLED; n++) {
			if (ledPosMed[n].x!=0 && ledPosMed[n].y!=0) {
				if (n==0) cairo_set_source_rgb(cr, 1,0,0);
				if (n==1) cairo_set_source_rgb(cr, 0,1,0);
				if (n==2) cairo_set_source_rgb(cr, 0,0,1);
				if (n==3) cairo_set_source_rgb(cr, 1,0,1);
				if (n==4) cairo_set_source_rgb(cr, 1,1,0);
				if (n==5) cairo_set_source_rgb(cr, 0,1,1);
				cairo_move_to(cr, 400, 600);
				cairo_line_to(cr, ledPosMed[n].x, ledPosMed[n].y);
				cairo_move_to(cr, 600, 600);
				cairo_line_to(cr, ledPosMed[n].x, ledPosMed[n].y);
				cairo_stroke(cr);
			}
		}

		cairo_set_line_width(cr, 5);
		cairo_set_source_rgb(cr, 0,0,0);
		cairo_move_to(cr, ccdXStart, 760);
		cairo_line_to(cr, ccdXEnd, 760);
		cairo_stroke(cr);

		cairoSdlBlit(cr);
		waitTouch(fd);
	}
}

int getCoords(BlackMark *meas, int no, touchedPointStruct *points) {
	int p1, p2, l, ccdpos;
	VectorStruct base1, base2, dir1, dir2, pt;
	int ret=0;
	int pointCnt=0;
	
	//Get intersection points between the lines from the shadow to the LED.
	for (p1=0; p1<no; p1++) {
		for (p2=p1+1; p2<no; p2++) {
			if (meas[p1].led!=meas[p2].led) {
				base1.x=ccdXStart+((ccdXEnd-ccdXStart)/(double)CCDMAX*(meas[p1].start+meas[p1].len/2));
				base1.y=ccdHeightAvg;
				base2.x=ccdXStart+((ccdXEnd-ccdXStart)/(double)CCDMAX*(meas[p2].start+meas[p2].len/2));
				base2.y=ccdHeightAvg;
				dir1.x=base1.x-ledPosMed[meas[p1].led].x;
				dir1.y=base1.y-ledPosMed[meas[p1].led].y;
				dir2.x=base2.x-ledPosMed[meas[p2].led].x;
				dir2.y=base2.y-ledPosMed[meas[p2].led].y;
				getIntersectPoint(base1, dir1, base2, dir2, &pt);
				points[pointCnt].x=pt.x;
				points[pointCnt].y=pt.y;
				points[pointCnt].confidence=1;
				points[pointCnt].size=(meas[p1].len+meas[p2].len)/2;
				points[pointCnt].size=abs(points[pointCnt].size*(ccdXEnd-ccdXStart))/CCDMAX;
				pointCnt++;
			}
		}
	}

	//Filter out 'ghosts': points where two shadow-led-lines intersect but where a third LED doesn't
	//see a shadow.
	for (p1=0; p1<pointCnt; p1++) {
		for (l=0; l<MAXLED; l++) {
			if (ledPosMed[l].x!=0 || ledPosMed[l].y!=0) {
				//See if we either have no chance of finding a shadow for this led (because it's not in range of
				//the sensor bar) or if we do, if there is one.
				base1.x=ledPosMed[l].x; base1.y=ledPosMed[l].y;
				dir1.x=points[p1].x-ledPosMed[l].x; dir1.y=points[p1].y-ledPosMed[l].y;
				base2.x=0; base2.y=ccdHeightAvg;
				dir2.x=1; dir2.y=0;
				getIntersectPoint(base1, dir1, base2, dir2, &pt); //pt is where we're supposed to find the shadow
				ccdpos=(pt.x-ccdXStart)*CCDMAX/(ccdXEnd-ccdXStart);
				if (ccdpos>0 && ccdpos<CCDMAX) {
					//We should find a shadow cast on the CCD. If not, disregard this point.
					int foundSpot=0;
					for (p2=0; p2<no; p2++) {
						int minoverlap=0;//meas[p2].len*.3;
						if (meas[p2].led==l && ccdpos>meas[p2].start+minoverlap && ccdpos<(meas[p2].start+meas[p2].len-minoverlap)) {
							foundSpot=1;
//							printf("Found shadow of spot %i from led %i at ccdpos %i.\n", p1, l, ccdpos);
						}
					}
					if (!foundSpot) {
//						printf("Expecting led %i spot shadow at %i; not found!\n", l, ccdpos);
						points[p1].confidence=CONFIDENCE_GHOST;
					}	
				}
			}
		}
	}

	//Join overlapping points. They usually are the result of the same finger/pen detected by different
	//LEDs.
	for (p1=0; p1<pointCnt; p1++) {
		for (p2=p1+1; p2<pointCnt; p2++) {
			if (points[p2].confidence>0 && points[p1].confidence>0) {
				int distance=sqrt(pow(points[p1].x-points[p2].x, 2)+pow(points[p1].y-points[p2].y, 2));
				if (distance<(points[p1].size/2) || distance<(points[p2].size/2)) {
					points[p1].x=(points[p1].x+points[p2].x)/2;
					points[p1].y=(points[p1].y+points[p2].y)/2;
					points[p1].confidence++;
					points[p1].size=(points[p1].size+points[p2].size)/2;
					points[p2].confidence=CONFIDENCE_DUPLICATE;
				}
			}
		}
	}

	//Get rid of unlikely points; e.g. when 3 points with a confidence of 3 appear, the fourth one with a
	//confidence of just one probably is a false reading.
	int somewhatConfidentCount=0;
	int totalPointCount=0;
	for (p1=0; p1<pointCnt; p1++) {
		if (points[p1].confidence>0) totalPointCount++;
		if (points[p1].confidence>1) somewhatConfidentCount++;
	}
	if (somewhatConfidentCount>=2 && totalPointCount>=2) {
		for (p1=0; p1<pointCnt; p1++) {
			if (points[p1].confidence>0 && points[p1].confidence<2) points[p1].confidence=CONFIDENCE_UNLIKELY;
		}
	}

	ret=pointCnt;
	return ret;
}
